/*
 * Copyright (C) 2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "para.h"
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h> /* open */
#include <sys/mman.h> /* mmap */
#include <unistd.h> /* fork */
#include <signal.h>
#include <sys/signal.h>





#include <string.h>

#include <SDL/SDL.h>
#include "SFont.h"

#include <SDL/SDL_image.h>
#include <SDL/SDL_events.h>
#include <SDL/SDL_keyboard.h>
#include <SDL/SDL_keysym.h>
#include <sys/time.h> /* timeval, needed for select */

#include "sdl_gui.cmdline.h"

//#define PIC_DIR "/usr/local/share/icc/pics" /* Really needed? */
#define DEFAULT_PIC "default.jpg"

#define FRAME_RED 100
#define FRAME_GREEN 200
#define FRAME_BLUE 200

SDL_Surface *screen;
static int width = 0;
static int height = 0;

struct gengetopt_args_info args_info;

#define FRAME_WIDTH 10
#define FONT_HEIGHT 36

#define INPUT_WIDTH width - 2 * FRAME_WIDTH
#define INPUT_X FRAME_WIDTH
#define INPUT_HEIGHT (args_info.interactive_flag? FONT_HEIGHT : 0)
#define INPUT_Y height - FRAME_WIDTH - INPUT_HEIGHT 

#define OUTPUT_WIDTH width - 2 * FRAME_WIDTH
#define OUTPUT_X FRAME_WIDTH
#define OUTPUT_Y FRAME_WIDTH 
#define OUTPUT_HEIGHT (height - INPUT_HEIGHT - 2 * FRAME_WIDTH)

#define NUM_LINES (height - 2 * FRAME_WIDTH - INPUT_HEIGHT) / FONT_HEIGHT

#define M_YELLOW 0
#define N_YELLOW 1
#define C_BLUE 2

#define LEFT 1
#define RIGHT 2
#define CENTER 3


struct stat_item{
        char name[MAXLINE];
        char prefix[MAXLINE];
        char postfix[MAXLINE];
        char content[MAXLINE];
        unsigned x;
        unsigned y;
        unsigned w;
        unsigned h;
	Uint8 r;
	Uint8 g;
	Uint8 b;
        int font;
        int align;
};

struct font {
	char name[MAXLINE];
        SFont_FontInfo fontinfo;
};

struct font fonts[] = {
	{"24P_Arial_Metallic_Yellow.png", },
	{"24P_Arial_NeonYellow.png",},
	{"24P_Copperplate_Blue.png",},
	{"", },
};


#define PIC_WIDTH width * 3 / 10
#define PIC_HEIGHT height * 2 / 5

#define INPUT_FONT 0
#define OUTPUT_FONT 1
#define MSG_FONT 2


enum {
	STATUS_BAR, 
	STATUS, 
	STATUS_FLAGS, 
	NUM_PLAYED, 
	SCORE, 
	STREAM, 
	FORMAT,
	LENGTH_MIN, 
	LENGTH_SEC,
	MTIME, 
	FILE_SIZE, 
	AUDIO_FILE_INFO1,
	AUDIO_FILE_INFO2,
	AUDIO_FILE_INFO3,
	DBINFO1, 
	DBINFO2,
	DBINFO3
};


struct stat_item stat_items_pic[] = {
        {
		.name = "status_bar:",
		.prefix = "",
		.postfix = "", 
		.content = "",
		.x = 0,
		.y = 10,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = M_YELLOW,
		.align = CENTER
        },
        {
		.name = "status:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 35,
		.y = 24,
		.w = 15,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = LEFT
        },
        {
		.name = "status_flags:",
		.prefix = " (",
		.postfix = ")",
		.content = "",
		.x = 50,
		.y = 24,
		.w = 10,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER
        },
  
      {
		.name = "num_played:",
		.prefix = "#",
		.postfix = "",
		.content = "0",
		.x = 60,
		.y = 24,
		.w = 15,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER
        },
      {
		.name = "score:",
		.prefix = "Score: ",
		.postfix = "",
		.content = "",
		.x = 75,
		.y = 24,
		.w = 25,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = RIGHT
        },
        {
		.name = "stream:",
		.prefix = "Stream: ",
		.postfix = "",
		.content = "no content yet",
		.x = 35,
		.y = 48,
		.w = 35,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = LEFT
        },
	{
		.name = "format:",
		.prefix = "Format: ",
		.postfix = "",
		.content = "",
		.x = 70,
		.y = 48,
		.w = 30,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = RIGHT
	},
        {
		.name = "length_min:",
		.prefix = "Length: ",
		.postfix = ":",
		.content = "empty",
		.x = 70,
		.y = 40,
		.w = 25,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = RIGHT
        },
        {
		.name = "length_sec:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 95,
		.y = 40,
		.w = 5,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = LEFT
        },
	{
		.name = "mtime:",
		.prefix = "MTime: ",
		.postfix = "",
		.content = "",
		.x = 35,
		.y = 32,
		.w = 65,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = LEFT
	},
	{
		.name = "file_size:",
		.prefix = "Size: ",
		.postfix = "kb",
		.content = "",
		.x = 35,
		.y = 40,
		.w = 35,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = LEFT
	},
	{
		.name = "audio_file_info1:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 60,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER
	}, 
	{
		.name = "audio_file_info2:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 65,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER 
	}, 
	{
		.name = "audio_file_info3:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 70,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER 
	}, 
	{
		.name = "dbinfo1:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 83,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER
	}, 
	{
		.name = "dbinfo2:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 88,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER 
	}, 
	{
		.name = "dbinfo3:",
		.prefix = "",
		.postfix = "",
		.content = "",
		.x = 0,
		.y = 93,
		.w = 100,
		.h = FONT_HEIGHT,
		.r = 0,
		.g = 0,
		.b = 0,
		.font = N_YELLOW,
		.align = CENTER 
	}, 
	{
		.name = ""
	}

};

struct stat_item *stat_items = stat_items_pic;

/*
 * init SDL libary and set window title
 */
static void init_SDL(void)
{
	if (SDL_Init(SDL_INIT_VIDEO) == -1) {
		fprintf(stderr,
			"Couldn't initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}
	/* Clean up on exit */
	atexit(SDL_Quit);
	/* Initialize the display */
	if (args_info.fullscreen_flag)
		screen = SDL_SetVideoMode(width, height, 0, SDL_FULLSCREEN);
	else
		screen = SDL_SetVideoMode(width, height, 0, 0);
	if (!screen) {
		fprintf(stderr, "Couldn't set video mode: %s\n", 
			SDL_GetError());
		exit(1);
	}
	SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
	/* Set the window manager title bar */
	SDL_WM_SetCaption("The Gui of death that makes you blind (paraslash " 
		VERSION ")", "SFont");
}

/*
 * draw rectangular frame of width FRAME_WIDTH
 */
static void draw_frame(Uint8 r, Uint8 g, Uint8 b) {
	SDL_Rect rect;

	rect.x = 0;
	rect.y = 0;
	rect.w = width;
	rect.h = FRAME_WIDTH;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, r, g, b));
	SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);

	rect.x = 0;
	rect.y = height - FRAME_WIDTH;
	rect.w = width;
	rect.h = FRAME_WIDTH;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, r, g, b));
	SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);

	rect.x = 0;
	rect.y = FRAME_WIDTH;
	rect.w = FRAME_WIDTH;
	rect.h = height -  2 * FRAME_WIDTH;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, r, g, b));
	SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);

	rect.x = width - FRAME_WIDTH;
	rect.y = FRAME_WIDTH;
	rect.w = FRAME_WIDTH;
	rect.h = height -  2 * FRAME_WIDTH;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, r, g, b));
	SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
}

/*
 * fill input rect with color
 */
static void fill_input_rect(void)
{
	SDL_Rect rect;

	rect.x = INPUT_X;
	rect.y = INPUT_Y;
	rect.w = INPUT_WIDTH;
	rect.h = INPUT_HEIGHT;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, 10, 150, 10));
}

/*
 * fill output rect with color
 */
static void fill_output_rect(void)
{
	SDL_Rect rect;

	rect.x = OUTPUT_X;
	rect.y = OUTPUT_Y;
	rect.w = OUTPUT_WIDTH;
	rect.h = OUTPUT_HEIGHT;
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, 0, 0, 0));
}

/*
 * convert tab to space
 */
static void tab2space(char *text)
{
	char *p = text;
	while (*p) {
		if (*p == '\t')
			*p = ' ';
		p++;
	}
}

static void print_msg(char *msg)
{
	SFont_FontInfo *font = &(fonts[MSG_FONT].fontinfo);
	char *buf = strdup(msg);
	int len = strlen(buf);
	
	if (!buf)
		return;
	while (TextWidth2(font, buf) > INPUT_WIDTH && len > 0) {
		*(buf + len) = '\0';
		len--;
	}
	fill_input_rect();
	PutString2(screen, font, INPUT_X, INPUT_Y, buf);
	free(buf);
}

static void update_all(void)
{
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

static void update_input(void)
{
	SDL_UpdateRect(screen, INPUT_X, INPUT_Y, INPUT_WIDTH, INPUT_HEIGHT);

}

/*
 * wait for key, ignore all other events, return 0 if there is no key event
 * pending. Otherwise return keysym of key
 */
SDLKey get_key(void)
{
	SDL_Event event;
	
	while (SDL_PollEvent(&event) > 0) {
		if(event.type != SDL_KEYDOWN)
			continue;
//		printf("Key pressed, scancode: 0x%x\n", 
//			event.key.keysym.scancode);
		return event.key.keysym.sym;
	}
	return 0;
}

/*
 * print message, wait for key (blocking), return 1 for 'q', 0 else
 */
static SDLKey hit_key(char *msg)
{
	SDLKey sym;

	print_msg(msg);
	update_input();
	while (!(sym = get_key()))
		;
	fill_input_rect();
	update_input();
	if (sym == SDLK_q)  
		return 1;
	else
		return 0;
}

/*
 * read paraslash command from input, execute it and print results
 */
static int command_handler(void)
{
	FILE *pipe;
	unsigned count = 0;
	char text[MAXLINE]="";
	char buf[MAXLINE]="";
	SFont_FontInfo *font = &fonts[OUTPUT_FONT].fontinfo;

//	printf("string input\n");
	SFont_Input2(screen, &fonts[INPUT_FONT].fontinfo, 
		INPUT_X, INPUT_Y - 5, INPUT_WIDTH, text);
	if (!strlen(text)) 
		return 1;
	if (!strcmp(text, "exit") || !strcmp(text, "quit"))
		return 0;
	if (text[0] == '!') {
		if (text[1] == '\0')
			return 1;
		pipe = popen(text + 1, "r");
	} else {
		sprintf(buf, BINDIR "/para_client %s 2>&1", text);
		pipe = popen(buf, "r");
	}
	if (!pipe)
		return 0;
	fill_output_rect();
	while(fgets(text, MAXLINE - 1, pipe)) {
		int len;

		tab2space(text);
		len = strlen(text);
	//	printf("string: %s\n", dest);
		while (TextWidth2(font, text) > width - 2 * FRAME_WIDTH && 
				len > 0) {
			text[len] = '\0';
			len--;
		}
		PutString2(screen, font, OUTPUT_X, 
			OUTPUT_Y + count * FONT_HEIGHT, text);
		count++; 
		if (count >= NUM_LINES) {
			update_all();
			if (hit_key("Hit any key to continue, q to return"))	
				goto out;
			count = 0;
			fill_output_rect();
		}
	}
	update_all();
	hit_key("Hit any key to return");
out:	fill_output_rect();
	pclose(pipe);
	return 1;
}
/*
 * Add prefix and postfix to string, delete characters from the end
 * if its length exceeds the max length defined in stat_items[item]
 */
static void transform_string(int item, char *dest)
{
	unsigned len = strlen(stat_items[item].name);
	char *content = stat_items[item].content;
	unsigned pixels = stat_items[item].w * (width - 2 * FRAME_WIDTH) / 100;
	SFont_FontInfo *font = &(fonts[stat_items[item].font].fontinfo);

	snprintf(dest, MAXLINE - 1 , "%s%s%s", stat_items[item].prefix, 
		content,  stat_items[item].postfix);
	len = strlen(dest);
//	printf("string: %s\n", dest);
	while (TextWidth2(font, dest) > pixels && len > 0) {
		*(dest + len) = '\0';
		len--;
	}
//	printf("trans string: %s\n", dest);
}

SDL_Surface *load_jpg()
{
	SDL_RWops *rwop;
	static FILE *pipe;
	char *pic_cmd[] = {"pic", NULL};

	if (pipe)
		pclose(pipe);
	if (!(pipe = popen_read_client(pic_cmd)))
		return NULL;
	if (!(rwop = SDL_RWFromFP(pipe, 0)))
		return NULL;
//	rwop = SDL_RWFromFile(path, "rb");
	return IMG_LoadJPG_RW(rwop);
}

void update_pic(void)
{
	SDL_Surface *img;
	SDL_Rect src_pic_rect = {
		.x = 0,
		.y = 0,
		.w = PIC_WIDTH,
		.h = PIC_HEIGHT, 
	};
	SDL_Rect dest_pic_rect = {
		.x = FRAME_WIDTH,
		.y = OUTPUT_HEIGHT / 5,
		.w = PIC_WIDTH,
		.h = PIC_HEIGHT, 
	};

	if (!screen)
		return;

	if (!(img = load_jpg()))
		return;
	SDL_FillRect(screen, &dest_pic_rect, SDL_MapRGB(screen->format, 
		0, 0, 0));
	SDL_BlitSurface(img, &src_pic_rect, screen, &dest_pic_rect);
	SDL_Flip(screen);
	SDL_FreeSurface(img);
}

/* 
 * update status item number i.
 */
static void do_update(int i)
{
	static int last_played = -1;
	SDL_Rect rect;
	char buf[MAXLINE];
	SFont_FontInfo *font = &(fonts[stat_items[i].font].fontinfo);

	rect.x = stat_items[i].x * (width - FRAME_WIDTH * 2) / 100 
		+ FRAME_WIDTH;
	rect.y = stat_items[i].y * (height - 2 * FRAME_WIDTH - INPUT_HEIGHT) 
		/ 100;
	rect.w = stat_items[i].w * (width - 2 * FRAME_WIDTH) / 100;
	rect.h = stat_items[i].h;
	transform_string(i, buf);
	SDL_FillRect(screen, &rect, SDL_MapRGB(screen->format, 
		stat_items[i].r, stat_items[i].g, stat_items[i].b));
	switch(stat_items[i].align) {
	case CENTER:
		PutString2(screen, font, 
			rect.x + (rect.w - TextWidth2(font, buf)) / 2, 
			rect.y, buf);
		break;
	case LEFT:
		PutString2(screen, font, rect.x, rect.y, buf);
		break;
	case RIGHT:
		PutString2(screen, font, rect.x + (rect.w - 
			TextWidth2(font, buf)), rect.y, buf);
		break;
	}
	SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
	if (i == NUM_PLAYED && atoi(stat_items[i].content) != last_played) {
		update_pic();
		last_played = atoi(stat_items[i].content);
	};
}

/*
 * Check if buf is a known status line. If so call do_update and return 1.
 * Return 0 otherwise.
 */
void update_status(char *buf)
{
	int i;
	unsigned len;
	
//	printf("Checking %s\n", buf);
	for (i = 0; (len = strlen(stat_items[i].name)); i++) {
		if (!strncmp(stat_items[i].name, buf, len)) {
			strcpy(stat_items[i].content, buf + len);
			do_update(i);
//			printf("updated #%d: %s\n", i, stat_items[i].content);
			return;
		} 
	}
}

/*
 * Read stat line from pipe if pipe is ready, call update_status to
 * display information. 
 */
int draw_status(FILE *pipe)
{
	static char buf[STRINGSIZE];
	static int offset;
	fd_set rfds;
	int file_num = fileno(pipe);
	int ret, n, file_no = fileno(pipe);
	struct timeval tv;
	
	tv.tv_sec = 0;
	tv.tv_usec = 300000;
	FD_ZERO(&rfds);
	FD_SET(file_num, &rfds);
	ret = select(file_num + 1, &rfds, NULL, NULL, &tv);
//	printf("select returned %d\n", ret);
	if (ret <= 0)
		return 0;
	
//	if (!fgets(buf, MAXLINE - 1, pipe))
//		return -1;
	n = read(file_no, buf + offset, sizeof(buf) - offset);
	if (n <= 0)
		return 0;	
	offset = check_buf_for_items(buf, n, &update_status, 0);
//	printf("Read: %s\n", buf);
//	buf[strlen(buf) -  1] = '\0';
//	update_status(buf);
	return 1;
}	

static void clean_exit(int ret)
{
	SDL_Quit();
	exit(ret);
}

static void print_help(void)
{
	print_msg("Hit q to quit, any other key to enter command mode");
}

static int configfile_exists(void)
{
	int cf_given = args_info.config_file_given;
	const char *default_fn = ".paraslash/sdl_gui.conf";
	static char cfa[_POSIX_PATH_MAX];
	
	if (cf_given)
		return file_exists(args_info.config_file_arg);
	sprintf(cfa, "%s/%s", getenv("HOME"), default_fn);
	args_info.config_file_arg = cfa;
	return file_exists(cfa);
}

/*
 * MAIN
 */
int main(int argc, char *argv[])
{
	int i, ret;
	FILE *pipe;
	SDLKey sym;

	cmdline_parser(argc, argv, &args_info);
	ret = configfile_exists();
//	printf("w=%i,h=%i,ret=%i, cf=%s\n", width, height, ret, args_info.config_file_arg);

	if (!ret && args_info.config_file_given) {
		fprintf(stderr, "Can't read config file %s\n", 
			args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (ret)
		cmdline_parser_configfile(args_info.config_file_arg, 
			&args_info, 0, 0, 0); 

	width = args_info.width_arg;
	height = args_info.height_arg;
//	printf("w=%i,h=%i,ret=%i, cf=%s\n", width, height, ret, args_info.config_file_arg);

	pipe = open_stat_pipe(); 
	init_SDL();
	for (i = 0; fonts[i].name[0]; i++) {
		char buf[MAXLINE];
		sprintf(buf, "%s/%s", FONTDIR, fonts[i].name);
		/* Load the font - You don't have to use the IMGlib for this */
		fonts[i].fontinfo.Surface = IMG_Load(buf); 
		/* Prepare the font for use */
		InitFont2(&fonts[i].fontinfo);
	}
	draw_frame(FRAME_RED, FRAME_GREEN, FRAME_BLUE);
	if (args_info.interactive_flag) {
		print_help();
		update_input();
	}
	for (;;) {
		ret = draw_status(pipe);
		if (ret < 0) {
			pclose(pipe);
			pipe = NULL;
		}
		if (SDL_QuitRequested())
			clean_exit(0);
		while ((sym = get_key())) {
			if (!args_info.interactive_flag)
				clean_exit(0);
			if (sym == SDLK_q)
				clean_exit(0);
			if (sym ==        SDLK_LSHIFT 
				|| sym == SDLK_RSHIFT 
				|| sym == SDLK_LMETA
				|| sym == SDLK_RMETA
				|| sym == SDLK_RCTRL
				|| sym == SDLK_LCTRL
				|| sym == SDLK_MODE
				|| sym == SDLK_CAPSLOCK
				|| sym == SDLK_LALT
				|| sym == SDLK_RALT
				|| sym == SDLK_RSUPER
				|| sym == SDLK_LSUPER
				|| sym == SDLK_COMPOSE
				)
				continue;
			if (pipe) {
//				printf("closing pipe\n");
		                kill(0, SIGINT);
				pclose(pipe);
//				printf("pipe closed\n");
			}
			fill_input_rect();
			update_input();
			if (!command_handler())
				clean_exit(0);
			fill_output_rect();
			print_help();
			update_pic();
			SDL_UpdateRect(screen, 0, 0, 0, 0);
			pipe = open_stat_pipe();
			break;
		}
	}
	return 0;
}
